from django.conf import settings
from django.contrib.messages.constants import DEFAULT_TAGS
from django.utils import translation

import graphene
from graphene import ObjectType
from graphene_django import DjangoObjectType
from graphene_django.forms.mutation import DjangoModelFormMutation
from haystack.inputs import AutoQuery
from haystack.query import SearchQuerySet
from haystack.utils.loading import UnifiedIndex

from .forms import PersonForm
from .models import Group, Notification, Person, TaskUserAssignment
from .util.core_helpers import get_allowed_object_ids, get_app_module, get_app_packages, has_person
from .util.frontend_helpers import get_language_cookie


class NotificationType(DjangoObjectType):
    class Meta:
        model = Notification


class PersonType(DjangoObjectType):
    class Meta:
        model = Person

    full_name = graphene.Field(graphene.String)

    def resolve_full_name(root: Person, info, **kwargs):
        return root.full_name


class GroupType(DjangoObjectType):
    class Meta:
        model = Group


class LanguageType(ObjectType):
    code = graphene.String(required=True)
    name = graphene.String(required=True)
    name_local = graphene.String(required=True)
    name_translated = graphene.String(required=True)
    bidi = graphene.Boolean(required=True)
    cookie = graphene.String(required=True)


class SystemPropertiesType(graphene.ObjectType):
    current_language = graphene.String(required=True)
    available_languages = graphene.List(LanguageType)

    def resolve_current_language(parent, info, **kwargs):
        return info.context.LANGUAGE_CODE

    def resolve_available_languages(parent, info, **kwargs):
        return [
            translation.get_language_info(code) | {"cookie": get_language_cookie(code)}
            for code, name in settings.LANGUAGES
        ]


class CeleryProgressMessage(ObjectType):
    message = graphene.String(required=True)
    level = graphene.Int(required=True)
    tag = graphene.String(required=True)

    def resolve_message(root, info, **kwargs):
        return root[1]

    def resolve_level(root, info, **kwargs):
        return root[0]

    def resolve_tag(root, info, **kwargs):
        return DEFAULT_TAGS.get(root[0], "info")


class CeleryProgressAdditionalButtonType(ObjectType):
    title = graphene.String(required=True)
    url = graphene.String(required=True)
    icon = graphene.String()


class CeleryProgressMetaType(DjangoObjectType):
    additional_button = graphene.Field(CeleryProgressAdditionalButtonType, required=False)

    class Meta:
        model = TaskUserAssignment
        fields = (
            "title",
            "back_url",
            "progress_title",
            "error_message",
            "success_message",
            "redirect_on_success_url",
            "additional_button",
        )

    def resolve_additional_button(root, info, **kwargs):
        if not root.additional_button_title or not root.additional_button_url:
            return None
        return {
            "title": root.additional_button_title,
            "url": root.additional_button_url,
            "icon": root.additional_button_icon,
        }


class CeleryProgressProgressType(ObjectType):
    current = graphene.Int()
    total = graphene.Int()
    percent = graphene.Float()


class CeleryProgressType(graphene.ObjectType):
    state = graphene.String()
    complete = graphene.Boolean()
    success = graphene.Boolean()
    progress = graphene.Field(CeleryProgressProgressType)
    messages = graphene.List(CeleryProgressMessage)
    meta = graphene.Field(CeleryProgressMetaType)

    def resolve_messages(root, info, **kwargs):  # noqa
        if root["complete"] and isinstance(root["result"], list):
            return root["result"]
        return root["progress"].get("messages", [])


class SearchModelType(graphene.ObjectType):
    absolute_url = graphene.String()
    name = graphene.String()
    icon = graphene.String()

    def resolve_absolute_url(root, info, **kwargs):
        if hasattr(root, "get_absolute_url"):
            return root.get_absolute_url()
        else:
            return "#!"

    def resolve_name(root, info, **kwargs):
        return str(root)

    def resolve_icon(root, info, **kwargs):
        return getattr(root, "icon_", "")


class SearchResultType(graphene.ObjectType):
    app_label = graphene.String()
    model_name = graphene.String()
    score = graphene.Int()
    obj = graphene.Field(SearchModelType)
    verbose_name = graphene.String()
    verbose_name_plural = graphene.String()
    text = graphene.String()

    def resolve_obj(root, info, **kwargs):  # noqa
        return root.object


class PersonMutation(DjangoModelFormMutation):
    person = graphene.Field(PersonType)

    class Meta:
        form_class = PersonForm


class MarkNotificationReadMutation(graphene.Mutation):
    class Arguments:
        id = graphene.ID()  # noqa

    notification = graphene.Field(NotificationType)

    @classmethod
    def mutate(cls, root, info, id):  # noqa
        notification = Notification.objects.get(pk=id)
        # FIXME permissions
        notification.read = True
        notification.save()

        return notification


class Query(graphene.ObjectType):
    ping = graphene.String(default_value="pong")

    notifications = graphene.List(NotificationType)

    persons = graphene.List(PersonType)
    person_by_id = graphene.Field(PersonType, id=graphene.ID())
    who_am_i = graphene.Field(PersonType)

    system_properties = graphene.Field(SystemPropertiesType)

    celery_progress_by_task_id = graphene.Field(CeleryProgressType, task_id=graphene.String())

    search_snippets = graphene.List(
        SearchResultType, query=graphene.String(), limit=graphene.Int(required=False)
    )

    def resolve_notifications(root, info, **kwargs):
        # FIXME do permission stuff
        return Notification.objects.all()

    def resolve_persons(root, info, **kwargs):
        # FIXME do permission stuff
        return Person.objects.all()

    def resolve_person_by_id(root, info, id):  # noqa
        return Person.objects.get(pk=id)

    def resolve_who_am_i(root, info, **kwargs):
        if has_person(info.context.user):
            return info.context.user.person
        else:
            return None

    def resolve_system_properties(root, info, **kwargs):
        return True

    def resolve_celery_progress_by_task_id(root, info, task_id, **kwargs):
        print(TaskUserAssignment.objects.filter(task_result__task_id=task_id), "HDS")
        print("DOES DO ")
        print("RUN")
        task = TaskUserAssignment.objects.get(task_result__task_id=task_id)

        if not info.context.user.has_perm("core.view_progress_rule", task):
            return None
        progress = task.get_progress()
        progress["meta"] = task
        return progress

    def resolve_search_snippets(root, info, query, limit=-1, **kwargs):
        indexed_models = UnifiedIndex().get_indexed_models()
        allowed_object_ids = get_allowed_object_ids(info.context.user, indexed_models)
        results = SearchQuerySet().filter(id__in=allowed_object_ids).filter(text=AutoQuery(query))

        if limit < 0:
            return results

        return results[:limit]


class Mutation(graphene.ObjectType):
    update_person = PersonMutation.Field()

    mark_notification_read = MarkNotificationReadMutation.Field()


def build_global_schema():
    """Build global GraphQL schema from all apps."""
    query_bases = [Query]
    mutation_bases = [Mutation]

    for app in get_app_packages():
        schema_mod = get_app_module(app, "schema")
        if not schema_mod:
            # The app does not define a schema
            continue

        if AppQuery := getattr(schema_mod, "Query", None):
            query_bases.append(AppQuery)
        if AppMutation := getattr(schema_mod, "Mutation", None):
            mutation_bases.append(AppMutation)

    # Define classes using all query/mutation classes as mixins
    #  cf. https://docs.graphene-python.org/projects/django/en/latest/schema/#adding-to-the-schema
    GlobalQuery = type("GlobalQuery", tuple(query_bases), {})
    GlobalMutation = type("GlobalMutation", tuple(mutation_bases), {})

    return graphene.Schema(query=GlobalQuery, mutation=GlobalMutation)


schema = build_global_schema()
